# board.py
# 25.03.2012, N. Ramagnano

# original code from http://zetcode.com/wxpython/thetetrisgame/
# updated for dual player mode

import wx
import random
import threading


BOARD_LINES_REMOVED = wx.NewEventType()
BOARD_GAME_OVER = wx.NewEventType()
EVT_BOARD_LINES_REMOVED = wx.PyEventBinder(BOARD_LINES_REMOVED, 0)
EVT_BOARD_GAME_OVER = wx.PyEventBinder(BOARD_GAME_OVER, 0)


class BoardEvent(wx.PyCommandEvent):
    def __init__(self, eventType, windowID, data):
        wx.PyCommandEvent.__init__(self, eventType, windowID)
        self.data = data

    def Clone(self):
        self.__class__(self.GetId(), self.data)


class Board(wx.Panel):
    BoardWidth = 10
    BoardHeight = 22

    def __init__(self, parent, side):
        wx.Panel.__init__(self, parent)

        self.side = side
        
        self.isWaitingAfterLine = False
        self.curPiece = Shape()
        self.nextPiece = Shape()
        self.curX = 0
        self.curY = 0
        self.numLinesRemoved = 0
        self.board = []

        self.isStarted = False
        self.isPaused = False

        self.isGameOver = False
        self.isWinner = False

        self.Bind(wx.EVT_PAINT, self.OnPaint)
        
        self.SetDoubleBuffered(True)
        
        self.pixel_per_square_init = 20
        self.pixel_per_square = self.pixel_per_square_init
        self.SetMinSize((self.BoardWidth*self.pixel_per_square, self.BoardHeight*self.pixel_per_square))
        
        self.clearBoard()
        
        
    def smaller(self):
        if self.pixel_per_square > 15:
            self.set_board_size(self.pixel_per_square - 1)
        
        
    def bigger(self):
        frame_size = self.GetParent().GetSize()
        display_rect = wx.GetClientDisplayRect() #wx.GetDisplaySize()
        if frame_size[0] < (display_rect[2]-display_rect[0]-self.BoardWidth) and frame_size[1] < (display_rect[3]-display_rect[1]-self.BoardHeight):
            self.set_board_size(self.pixel_per_square + 1)
    
    def set_board_size(self, pixel_per_square):
        self.pixel_per_square = pixel_per_square
        self.SetMinSize((self.BoardWidth*self.pixel_per_square, \
                        self.BoardHeight*self.pixel_per_square))
        self.GetParent().GetSizer().SetSizeHints(self.GetParent())
        self.GetParent().Center()

    def shapeAt(self, x, y):
        return self.board[(y * Board.BoardWidth) + x]

    def setShapeAt(self, x, y, shape):
        self.board[(y * Board.BoardWidth) + x] = shape

    def squareWidth(self):
        return self.GetClientSize().GetWidth() / Board.BoardWidth

    def squareHeight(self):
        return self.GetClientSize().GetHeight() / Board.BoardHeight

    def clearBoard(self):
        self.board = []
        for i in range((Board.BoardHeight+4) * Board.BoardWidth):
            self.board.append(Tetrominoes.NoShape)

    def OnPaint(self, event):

        dc = wx.PaintDC(self)

        size = self.GetClientSize()
        boardTop = size.GetHeight() - Board.BoardHeight * self.squareHeight()
        
        # draw grid
        dc.SetPen(wx.TRANSPARENT_PEN)
        if self.side == "left":
            dc.SetBrush(wx.Brush('#ddeede'))
        else:
            dc.SetBrush(wx.Brush('#ddeeee'))
        for m in range(Board.BoardHeight):
            for n in range(Board.BoardWidth):
                if (m+n) % 2:
                    dc.DrawRectangle(n*self.squareWidth(), m*self.squareHeight(), self.squareWidth(), self.squareHeight())
        
        for i in range(Board.BoardHeight):
            for j in range(Board.BoardWidth):
                shape = self.shapeAt(j, Board.BoardHeight - i - 1)
                if shape != Tetrominoes.NoShape:
                    self.drawSquare(dc,
                        0 + j * self.squareWidth(),
                        boardTop + i * self.squareHeight(), shape)

        if self.curPiece.shape() != Tetrominoes.NoShape:
            for i in range(4):
                x = self.curX + self.curPiece.x(i)
                y = self.curY - self.curPiece.y(i)
                self.drawSquare(dc, 0 + x * self.squareWidth(),
                    boardTop + (Board.BoardHeight - y - 1) * self.squareHeight(),
                    self.curPiece.shape())
        
        if self.isPaused:
            self.drawBigText(dc, "Pause")
        else:
            if self.isWinner:
                self.drawBigText(dc, "Gewonnen")
            else:
                if self.isGameOver:
                    self.drawBigText(dc, "Verloren")

        if self.isStarted or self.isPaused or self.isWinner or self.isGameOver:
            self.drawScore(dc)

    def drawScore(self, dc):
        text = str(self.numLinesRemoved)
        font = wx.Font(self.squareWidth(), wx.FONTFAMILY_MODERN, wx.FONTSTYLE_NORMAL, wx.FONTWEIGHT_BOLD)
        dc.SetFont(font)
        text_size = dc.GetTextExtent(text)
        if self.side == 'left':
            dc.DrawText(text, self.squareWidth(), self.squareHeight())
        else:
            dc.DrawText(text, self.GetClientSize().GetWidth()-self.squareWidth()-text_size[0], self.squareHeight())

    def drawBigText(self, dc, text):
        font = wx.Font(self.squareWidth()*1.3, wx.FONTFAMILY_MODERN, wx.FONTSTYLE_NORMAL, wx.FONTWEIGHT_BOLD)
        dc.SetFont(font)
        text_size = dc.GetTextExtent(text)
        dc.DrawText(text, self.GetClientSize().GetWidth()/2-text_size[0]/2, self.GetClientSize().GetHeight()/2-text_size[1]/2)

    def dropDown(self):
        newY = self.curY
        while newY > 0:
            if not self.tryMove(self.curPiece, self.curX, newY - 1):
                break
            newY -= 1

        self.pieceDropped()

    def oneLineDown(self):
        if not self.tryMove(self.curPiece, self.curX, self.curY - 1):
            self.pieceDropped()
        


    def pieceDropped(self):
        for i in range(4):
            x = self.curX + self.curPiece.x(i)
            y = self.curY - self.curPiece.y(i)
            self.setShapeAt(x, y, self.curPiece.shape())

        self.removeFullLines()

        if not self.isWaitingAfterLine:
            self.newPiece()


    def removeFullLines(self):
        numFullLines = 0

        rowsToRemove = []

        for i in range(Board.BoardHeight):
            n = 0
            for j in range(Board.BoardWidth):
                if not self.shapeAt(j, i) == Tetrominoes.NoShape:
                    n = n + 1

            if n == self.BoardWidth:
                rowsToRemove.append(i)

        rowsToRemove.reverse()

        for m in rowsToRemove:
            for k in range(m, Board.BoardHeight-1):
                for l in range(Board.BoardWidth):
                    self.setShapeAt(l, k, self.shapeAt(l, k + 1))
        
        numFullLines = numFullLines + len(rowsToRemove)

        if numFullLines > 0:
            self.numLinesRemoved = self.numLinesRemoved + numFullLines
            
            event = BoardEvent(BOARD_LINES_REMOVED, self.GetId(), str(numFullLines))
            self.GetEventHandler().AddPendingEvent(event)
            
            self.isWaitingAfterLine = True
            self.curPiece.setShape(Tetrominoes.NoShape)
            self.Refresh()


    def newPiece(self):
        self.curPiece = self.nextPiece
        self.nextPiece.setRandomShape()
        self.curX = Board.BoardWidth / 2
        self.curY = Board.BoardHeight - 1 + self.curPiece.minY()

        if not self.tryMove(self.curPiece, self.curX, self.curY):
            self.curPiece.setShape(Tetrominoes.NoShape)
            self.isStarted = False
            self.isGameOver = True
            event = BoardEvent(BOARD_GAME_OVER, self.GetId(), 'Game over')
            self.GetEventHandler().AddPendingEvent(event)
        
        self.Refresh()

    def tryMove(self, newPiece, newX, newY):
        for i in range(4):
            x = newX + newPiece.x(i)
            y = newY - newPiece.y(i)
            if x < 0 or x >= Board.BoardWidth or y < 0 or y >= Board.BoardHeight:
                return False
            if self.shapeAt(x, y) != Tetrominoes.NoShape:
                return False

        self.curPiece = newPiece
        self.curX = newX
        self.curY = newY
        self.Refresh()
        return True

    def drawSquare(self, dc, x, y, shape):
        colors = ['#000000', '#CC6666', '#66CC66', '#6666CC',
                  '#CCCC66', '#CC66CC', '#66CCCC', '#DAAA00']

        light = ['#000000', '#F89FAB', '#79FC79', '#7979FC', 
                 '#FCFC79', '#FC79FC', '#79FCFC', '#FCC600']

        dark = ['#000000', '#803C3B', '#3B803B', '#3B3B80', 
                 '#80803B', '#803B80', '#3B8080', '#806200']

        pen = wx.Pen(light[shape])
        pen.SetCap(wx.CAP_PROJECTING)
        dc.SetPen(pen)

        dc.DrawLine(x, y + self.squareHeight() - 1, x, y)
        dc.DrawLine(x, y, x + self.squareWidth() - 1, y)

        darkpen = wx.Pen(dark[shape])
        darkpen.SetCap(wx.CAP_PROJECTING)
        dc.SetPen(darkpen)

        dc.DrawLine(x + 1, y + self.squareHeight() - 1,
            x + self.squareWidth() - 1, y + self.squareHeight() - 1)
        dc.DrawLine(x + self.squareWidth() - 1, 
        y + self.squareHeight() - 1, x + self.squareWidth() - 1, y + 1)

        dc.SetPen(wx.TRANSPARENT_PEN)
        dc.SetBrush(wx.Brush(colors[shape]))
        dc.DrawRectangle(x + 1, y + 1, self.squareWidth() - 2, 
        self.squareHeight() - 2)


class Tetrominoes(object):
    NoShape = 0
    ZShape = 1
    SShape = 2
    LineShape = 3
    TShape = 4
    SquareShape = 5
    LShape = 6
    MirroredLShape = 7


class Shape(object):
    coordsTable = (
        ((0, 0),     (0, 0),     (0, 0),     (0, 0)),
        ((0, -1),    (0, 0),     (-1, 0),    (-1, 1)),
        ((0, -1),    (0, 0),     (1, 0),     (1, 1)),
        ((0, -1),    (0, 0),     (0, 1),     (0, 2)),
        ((-1, 0),    (0, 0),     (1, 0),     (0, 1)),
        ((0, 0),     (1, 0),     (0, 1),     (1, 1)),
        ((-1, -1),   (0, -1),    (0, 0),     (0, 1)),
        ((1, -1),    (0, -1),    (0, 0),     (0, 1))
    )

    def __init__(self):
        self.coords = [[0,0] for i in range(4)]
        self.pieceShape = Tetrominoes.NoShape

        self.setShape(Tetrominoes.NoShape)

    def shape(self):
        return self.pieceShape

    def setShape(self, shape):
        table = Shape.coordsTable[shape]
        for i in range(4):
            for j in range(2):
                self.coords[i][j] = table[i][j]

        self.pieceShape = shape

    def setRandomShape(self):
        self.setShape(random.randint(1, 7))

    def x(self, index):
        return self.coords[index][0]

    def y(self, index):
        return self.coords[index][1]

    def setX(self, index, x):
        self.coords[index][0] = x

    def setY(self, index, y):
        self.coords[index][1] = y

    def minX(self):
        m = self.coords[0][0]
        for i in range(4):
            m = min(m, self.coords[i][0])

        return m

    def maxX(self):
        m = self.coords[0][0]
        for i in range(4):
            m = max(m, self.coords[i][0])

        return m

    def minY(self):
        m = self.coords[0][1]
        for i in range(4):
            m = min(m, self.coords[i][1])

        return m

    def maxY(self):
        m = self.coords[0][1]
        for i in range(4):
            m = max(m, self.coords[i][1])

        return m

    def rotatedLeft(self):
        if self.pieceShape == Tetrominoes.SquareShape:
            return self

        result = Shape()
        result.pieceShape = self.pieceShape
        for i in range(4):
            result.setX(i, self.y(i))
            result.setY(i, -self.x(i))

        return result

    def rotatedRight(self):
        if self.pieceShape == Tetrominoes.SquareShape:
            return self

        result = Shape()
        result.pieceShape = self.pieceShape
        for i in range(4):
            result.setX(i, -self.y(i))
            result.setY(i, self.x(i))

        return result
